from gluonts.dataset.common import ListDataset
from gluonts.dataset.field_names import FieldName
from gluonts.model.deepar import DeepAREstimator
from gluonts.mx.distribution.neg_binomial import NegativeBinomialOutput
from gluonts.mx.distribution.gaussian import GaussianOutput
from gluonts.mx.trainer import Trainer
from gluonts.evaluation.backtest import make_evaluation_predictions
from tqdm import tqdm
from htsmodels.results.calculate_metrics import CalculateResultsBottomUp
import numpy as np
import pickle
from pathlib import Path
import time


class DeepAR:
    def __init__(
        self,
        dataset,
        groups,
        input_dir="./",
        n_samples=500,
        store_prediction_samples=False,
        store_prediction_points=False,
    ):
        self.dataset = dataset
        self.groups = groups
        self.n_samples = n_samples
        self.timer_start = time.time()
        self.wall_time_preprocess = None
        self.wall_time_build_model = None
        self.wall_time_train = None
        self.wall_time_predict = None
        self.wall_time_total = None
        self.input_dir = input_dir
        self._create_directories()
        self.stat_cat_cardinalities = [
            v for k, v in self.groups["train"]["groups_n"].items()
        ]
        self.stat_cat = np.concatenate(
            ([v.reshape(-1, 1) for k, v in self.groups["train"]["groups_idx"].items()]),
            axis=1,
        )
        self.dates = [groups["dates"][0] for _ in range(groups["train"]["s"])]
        self.store_prediction_samples = store_prediction_samples
        self.store_prediction_points = store_prediction_points

        time_interval = (self.groups["dates"][1] - self.groups["dates"][0]).days
        if time_interval < 8:
            self.time_int = "W"
        elif time_interval < 32:
            self.time_int = "M"
        elif time_interval < 93:
            self.time_int = "Q"
        elif time_interval < 367:
            self.time_int = "Y"

    def _create_directories(self):
        # Create directory to store results if does not exist
        Path(f"{self.input_dir}results").mkdir(parents=True, exist_ok=True)

    def _build_train_ds(self):
        train_target_values = self.groups["train"]["data"].T

        train_ds = ListDataset(
            [
                {
                    FieldName.TARGET: target,
                    FieldName.START: start,
                    FieldName.FEAT_STATIC_CAT: fsc,
                }
                for (target, start, fsc) in zip(
                    train_target_values, self.dates, self.stat_cat
                )
            ],
            freq=self.time_int,
        )

        return train_ds

    def _build_test_ds(self):
        test_target_values = self.groups["predict"]["data"].reshape(
            self.groups["predict"]["s"], self.groups["predict"]["n"]
        )

        test_ds = ListDataset(
            [
                {
                    FieldName.TARGET: target,
                    FieldName.START: start,
                    FieldName.FEAT_STATIC_CAT: fsc,
                }
                for (target, start, fsc) in zip(
                    test_target_values, self.dates, self.stat_cat
                )
            ],
            freq=self.time_int,
        )

        return test_ds

    def train(self, lr=1e-3, epochs=100, dist="Gaussian"):
        self.wall_time_preprocess = time.time() - self.timer_start
        train_ds = self._build_train_ds()
        self.wall_time_build_model = (
            time.time() - self.timer_start - self.wall_time_preprocess
        )

        if dist == "NegativeBinomial":
            distribution = NegativeBinomialOutput()
        else:
            distribution = GaussianOutput()

        estimator = DeepAREstimator(
            prediction_length=self.groups["h"],
            freq=self.time_int,
            distr_output=distribution,
            use_feat_dynamic_real=False,
            use_feat_static_cat=True,
            cardinality=self.stat_cat_cardinalities,
            trainer=Trainer(learning_rate=lr, epochs=epochs, num_batches_per_epoch=50),
        )

        model = estimator.train(train_ds)
        self.wall_time_train = (
            time.time() - self.timer_start - self.wall_time_build_model
        )
        return model

    def predict(self, model):
        test_ds = self._build_test_ds()

        forecast_it, ts_it = make_evaluation_predictions(
            dataset=test_ds, predictor=model, num_samples=self.n_samples
        )

        print("Obtaining time series predictions ...")
        forecasts = list(tqdm(forecast_it, total=len(test_ds)))
        self.wall_time_predict = time.time() - self.timer_start - self.wall_time_train

        return forecasts

    def results(self, forecasts):
        res = np.zeros((len(forecasts), self.n_samples, self.groups["h"]))
        for i, j in enumerate(forecasts):
            res[i] = j.samples

        res = np.transpose(
            np.concatenate(
                (
                    np.zeros(
                        (
                            self.groups["train"]["s"],
                            self.n_samples,
                            self.groups["train"]["n"],
                        ),
                        dtype=np.float64,
                    ),
                    res,
                ),
                axis=2,
            ),
            (2, 0, 1),
        )

        return res

    def store_metrics(self, res):
        with open(
            f"{self.input_dir}results/results_gp_cov_{self.dataset}.pickle", "wb"
        ) as handle:
            pickle.dump(res, handle, pickle.HIGHEST_PROTOCOL)

    def metrics(self, samples):
        calc_results = CalculateResultsBottomUp(
            samples,
            self.groups,
            self.store_prediction_samples,
            self.store_prediction_points,
        )
        res = calc_results.calculate_metrics()
        self.wall_time_total = time.time() - self.timer_start

        res["wall_time"] = {}
        res["wall_time"]["wall_time_preprocess"] = self.wall_time_preprocess
        res["wall_time"]["wall_time_build_model"] = self.wall_time_build_model
        res["wall_time"]["wall_time_train"] = self.wall_time_train
        res["wall_time"]["wall_time_predict"] = self.wall_time_predict
        res["wall_time"]["wall_time_total"] = self.wall_time_total

        return res



import unittest
import tsaugmentation as tsag
from htsmodels.models.deepar import DeepAR
import shutil


class TestModel(unittest.TestCase):

    def setUp(self):
        self.preproc = tsag.preprocessing.PreprocessDatasets('tourism', test_size=228*10)
        self.data = self.preproc._tourism()
        self.n = self.data['predict']['n']
        self.s = self.data['train']['s']
        self.deepar = DeepAR(dataset='tourism', groups=self.data,
                             store_prediction_samples=True,
                             store_prediction_points=True)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree("./results")

    def test_results_mean_and_prediction_interval(self):
        model = self.deepar.train(epochs=2)
        forecasts = self.deepar.predict(model)
        samples = self.deepar.results(forecasts)
        res = self.deepar.metrics(samples)

        # Test shape of results
        self.assertTrue(res['mase']['bottom_ind'].shape == (self.s, ))
        self.assertTrue(res['CRPS']['bottom_ind'].shape == (self.s, ))
        self.assertTrue(res['rmse']['bottom_ind'].shape == (self.s, ))

        # Test shape of predictions
        # Number of prediction samples for the bottom time series (n_points, n_series, n_samples)
        self.assertTrue(res['predictions']['samples']['bottom'].shape == (self.n, self.s, 500))
        self.assertTrue(res['predictions']['points']['bottom'].shape == (self.data['h'], self.s))
        # Number of prediction samples for the total time series
        self.assertTrue(res['predictions']['samples']['total'].shape == (self.n, 500))
        self.assertTrue(res['predictions']['points']['total'].shape == (self.data['h'],))
        # Test number of objects predicted and stored
        self.assertTrue(len(res['predictions']['samples']) == 12)
        self.assertTrue(len(res['predictions']['points']) == 12)

    def test_results_mean_and_prediction_interval_without_storing_results(self):
        self.deepar = DeepAR(dataset='tourism', groups=self.data,
                          store_prediction_samples=False,
                          store_prediction_points=False)
        model = self.deepar.train(epochs=2)
        forecasts = self.deepar.predict(model)
        samples = self.deepar.results(forecasts)
        res = self.deepar.metrics(samples)

        # Test shape of results
        self.assertTrue(res['mase']['bottom_ind'].shape == (self.s, ))
        self.assertTrue(res['CRPS']['bottom_ind'].shape == (self.s, ))
        self.assertTrue(res['rmse']['bottom_ind'].shape == (self.s, ))

        # Test shape of predictions
        # Test number of objects predicted and stored
        self.assertTrue(len(res) == 4)




import unittest
from htsmodels.models.standard_gp import SGP
import tsaugmentation as tsag
import shutil


class TestModel(unittest.TestCase):

    def setUp(self):
        self.data = tsag.preprocessing.PreprocessDatasets('prison').apply_preprocess()
        self.n = self.data['predict']['n']
        self.s = self.data['train']['s']
        shutil.rmtree("./data/original_datasets")
        self.gpf = SGP('prison', self.data)

    def test_results_interval_piecewise(self):
        model, like = self.gpf.train(n_iterations=500)
        samples = self.gpf.predict(model, like)
        res = self.gpf.metrics(samples)
        self.assertLess(res['mase']['bottom'], 2.5)

    def test_results_interval_linear(self):
        model, like = self.gpf.train(n_iterations=100, mean_function='linear')
        samples = self.gpf.predict(model, like)
        res = self.gpf.metrics(samples)
        self.assertLess(res['mase']['bottom'], 2.5)

    def test_results_interval_zero(self):
        model, like = self.gpf.train(n_iterations=100, mean_function='zero')
        samples = self.gpf.predict(model, like)
        res = self.gpf.metrics(samples)
        self.assertLess(res['mase']['bottom'], 4)

    def test_results_interval_zero_noclip(self):
        model, like = self.gpf.train(n_iterations=100, mean_function='zero')
        samples = self.gpf.predict(model, like, False)
        res = self.gpf.metrics(samples)
        self.assertLess(res['mase']['bottom'], 4)



import torch
import numpy as np
import gpytorch
from .gp import ExactGPModel
from .mean_functions import PiecewiseLinearMean, LinearMean, ZeroMean
from gpytorch.mlls import SumMarginalLogLikelihood
from htsmodels.results.calculate_metrics import CalculateResultsBottomUp
import pickle
import tsaugmentation as tsag
from pathlib import Path
import time


class SGP:

    def __init__(self, dataset, groups, input_dir='./', n_samples=500,
                 store_prediction_samples=False,
                 store_prediction_points=False):
        self.dataset = dataset
        self.groups = groups
        self.input_dir = input_dir
        self.timer_start = time.time()
        self.wall_time_preprocess = None
        self.wall_time_build_model = None
        self.wall_time_train = None
        self.wall_time_predict = None
        self.wall_time_total = None
        self.groups, self.dt = self._preprocess()
        self._create_directories()
        self.n_samples = n_samples
        self.store_prediction_samples = store_prediction_samples
        self.store_prediction_points = store_prediction_points

        self.train_x = torch.arange(groups['train']['n'])
        self.train_x = self.train_x.type(torch.DoubleTensor)
        self.train_x = self.train_x.unsqueeze(-1)
        self.train_y = torch.from_numpy(groups['train']['data'])

    def _create_directories(self):
        # Create directory to store results if does not exist
        Path(f'{self.input_dir}results').mkdir(parents=True, exist_ok=True)

    def _preprocess(self):
        dt = tsag.preprocessing.utils.DataTransform(self.groups)
        self.wall_time_preprocess = time.time() - self.timer_start
        return dt.std_transf_train(), dt

    def _build_cov_matrices(self):
        covs = []
        for i in range(self.groups['train']['s']):
            # RBF kernel
            rbf_kernel = gpytorch.kernels.RBFKernel()
            rbf_kernel.lengthscale = torch.tensor([1.])
            scale_rbf_kernel = gpytorch.kernels.ScaleKernel(rbf_kernel)
            scale_rbf_kernel.outputscale = torch.tensor([0.5])

            # Periodic Kernel
            periodic_kernel = gpytorch.kernels.PeriodicKernel()
            periodic_kernel.period_length = torch.tensor([self.groups['seasonality']])
            periodic_kernel.lengthscale = torch.tensor([0.5])
            scale_periodic_kernel = gpytorch.kernels.ScaleKernel(periodic_kernel)
            scale_periodic_kernel.outputscale = torch.tensor([1.5])

            # Cov Matrix
            cov = scale_rbf_kernel + scale_periodic_kernel
            covs.append(cov)

        return covs

    def _build_model(self, mean_function):
        covs = self._build_cov_matrices()
        n_changepoints = 4
        changepoints = np.linspace(0, self.groups['train']['n'], n_changepoints + 2)[1:-1]

        model_list = []
        likelihood_list = []
        for i in range(self.groups['train']['s']):
            likelihood_list.append(gpytorch.likelihoods.GaussianLikelihood())
            if mean_function == 'piecewise_linear':
                model_list.append(ExactGPModel(self.train_x,
                                               self.train_y[:, i],
                                               likelihood_list[i],
                                               covs[i],
                                               PiecewiseLinearMean(changepoints)))
            elif mean_function == 'zero':
                model_list.append(ExactGPModel(self.train_x,
                                               self.train_y[:, i],
                                               likelihood_list[i],
                                               covs[i],
                                               gpytorch.means.ZeroMean()))
            elif mean_function == 'linear':
                model_list.append(ExactGPModel(self.train_x,
                                               self.train_y[:, i],
                                               likelihood_list[i],
                                               covs[i],
                                               LinearMean(1)))

        self.wall_time_build_model = time.time() - self.timer_start - self.wall_time_preprocess
        return likelihood_list, model_list

    def train(self, n_iterations=500, lr=1e-3, mean_function='piecewise_linear'):
        likelihood_list, model_list = self._build_model(mean_function)

        model = gpytorch.models.IndependentModelList(*model_list)
        likelihood = gpytorch.likelihoods.LikelihoodList(*likelihood_list)

        mll = SumMarginalLogLikelihood(likelihood, model)

        model.train()
        likelihood.train()

        optimizer = torch.optim.Adam(model.parameters(), lr=lr)  # Includes GaussianLikelihood parameters

        for i in range(n_iterations):
            optimizer.zero_grad()
            output = model(*model.train_inputs)
            loss = -mll(output, model.train_targets)
            loss.backward()
            print('Iter %d/%d - Loss: %.3f' % (i + 1, n_iterations, loss.item()))
            optimizer.step()

        self.wall_time_train = time.time() - self.timer_start - self.wall_time_build_model
        return model, likelihood

    def predict(self, model, likelihood, clip=True):
        model.eval()
        likelihood.eval()

        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            test_x = torch.arange(self.groups['predict']['n']).type(torch.DoubleTensor)
            predictions = likelihood(*model(*[test_x for i in range(self.groups['predict']['s'])]))

        i = 0
        samples = np.zeros((self.n_samples, self.groups['predict']['n'], self.groups['predict']['s']))
        for pred in predictions:
            samples[:, :, i] = np.random.normal(pred.mean.detach().numpy(), np.sqrt(pred.variance.detach().numpy()),
                                                size=(self.n_samples, self.groups['predict']['n']))
            i += 1

        samples = np.transpose(samples, (1, 2, 0))

        # transform back the data
        samples = ((samples * self.dt.std_data[np.newaxis, :, np.newaxis]) + self.dt.mu_data[np.newaxis, :, np.newaxis])
        self.groups = self.dt.inv_transf_train()

        # Clip predictions to 0 if there are negative numbers
        if clip:
            samples[samples < 0] = 0

        self.wall_time_predict = time.time() - self.timer_start - self.wall_time_train
        return samples

    def store_metrics(self, res):
        with open(f'{self.input_dir}results/results_gp_cov_{self.dataset}.pickle', 'wb') as handle:
            pickle.dump(res, handle, pickle.HIGHEST_PROTOCOL)

    def metrics(self, samples):
        calc_results = CalculateResultsBottomUp(samples,
                                                self.groups,
                                                self.store_prediction_samples,
                                                self.store_prediction_points)
        res = calc_results.calculate_metrics()
        self.wall_time_total = time.time() - self.timer_start

        res['wall_time'] = {}
        res['wall_time']['wall_time_preprocess'] = self.wall_time_preprocess
        res['wall_time']['wall_time_build_model'] = self.wall_time_build_model
        res['wall_time']['wall_time_train'] = self.wall_time_train
        res['wall_time']['wall_time_predict'] = self.wall_time_predict
        res['wall_time']['wall_time_total'] = self.wall_time_total

        return res

